import java.awt.*;
import java.io.*;
import java.util.*;
import java.awt.image.*;
import javax.imageio.*;
import se.lth.cs.pt.window.DotWindow;
import se.lth.cs.pt.io.Keyboard;
import se.lth.cs.pt.window.Color;
import se.lth.cs.pt.window.GameEvent;


/**
 * Small convenience class, the standard library probably has something similar.
 * HS are my initals to prevent clashes with the library
 */
class HSPoint{
	
	double r, i;
	Color col;
	
	HSPoint(double rPos, double iPos, Color c){
		i = iPos;
		r = rPos;
		col = new Color(c.getRed(), c.getGreen(), c.getBlue());
	}
	
	HSPoint(){
		i = 0;
		r = 0;
		col = new Color(0,0,0);
	}
	
	HSPoint(double rPos, double iPos){
		i = iPos;
		r = rPos;
		col = new Color(0, 0, 0);
	}
	
	void setColor(Color c){
		col = new Color(c.getRed(), c.getGreen(), c.getBlue());
	}
	
	double calcAbsolute(){
		return Math.sqrt((r*r) + (i*i));
	}
}

/**
 * The fractal class that will contain all the points
 */
class HSFractal{
	
	double scale;		//the actual scale
	int size; 			//the size of the array (and window)
	HSPoint[][] fractal;//the array itself
	HSPoint start;		//bottom left corner
	int[] fPixels = null;
	
	HSFractal(int sz, double s, double start_r, double start_i){
		scale = s;
		size = sz;
		//The start point is the middle point - (half the window size divided by the scale)
		start = new HSPoint(start_r - (((double)size/2)/scale), start_i - (((double)size/2)/scale));
		fractal = new HSPoint[size][size];
	}
	
	//Here we do image saving work if the user chose to animate
	void saveImage(String outName){
		int i = 0;
		//1D-array where we save all the color info
		fPixels = new int[size*size];
		for(int re = 0; re < size; re++){
			for(int im = 0; im < size; im++){
				fPixels[i++] =  (255 << 24) |  (fractal[re][im].col.getRed() << 16) |  
				(fractal[re][im].col.getGreen() << 8 ) | fractal[re][im].col.getBlue();
			}
		}
		
		//our buffered image
		BufferedImage fBufferedImage = new BufferedImage(size, size, BufferedImage.TYPE_INT_ARGB);
		//the rasterization of the buffered image
		WritableRaster wr = fBufferedImage.getRaster();
		//the raster is filled with our colors
		wr.setDataElements (0, 0, size, size, fPixels);
		//the out file
		File out = new File (outName);
		try{
  			ImageIO.write (fBufferedImage, "png", out);
  		}catch(IOException e){
  			System.out.println(e.getMessage());
  		}
	}	
	
	HSPoint kvadrera(HSPoint p){
		//powering of complex numbers (kvadrera means to do a power of 2 in Swedish)
		return new HSPoint(Math.pow(p.r, 2) - Math.pow(p.i, 2), p.i * p.r * 2);
	}
	
	//Summing the number sequence
	int sumUp(HSPoint z, HSPoint c, int i){
		z = kvadrera(z);
		z.r += c.r;
		z.i += c.i;
		
		//We will keep at it recursively until we hit 255 iterations or get divergence
		if(i > 254 || z.calcAbsolute() > 2)
			return i;
		else
			return sumUp(z, c, ++i);
	}
	
	Color colorize(int v){
		//temporary variables	
		double red = 0; double green = 0; double blue = 0;
		//some nice colors
		if(v == 255){
			red=80; green=0; blue=0;
		}else if(v < 128){
			red=50; green=50; blue=(50+v)*1.44;
		}else{
			red=127-(v/2); green=100; blue=255-(2*(v-127)); 
		}
					
		return new Color((int)red, (int)green, (int)blue);
	}
	
	//here we give every point in the 2D-array it's values
	void calculateFractal(){
		for(int i = 0; i < size; i++){
			for(int j = 0; j < size; j++){
				//the coordinates are determined by starting with the start point and then adding the "pixel value" divided by the scale
				HSPoint temp = new HSPoint(start.r + (i/scale), start.i + (j/scale));
				//this is where we colorize
				temp.setColor(colorize(sumUp(temp, temp, 0)));
				fractal[i][j] = temp;
			}
		}
	}
}

/**
 * The class that handles the window
 */
class HSMandelbrotWin{
	
	double scale;			//the scale of the window
	int size;				//the side of the window
	double zoom;			//the factor of our zoom
	HSPoint animation;		//in case the user chooses to animate then this is the target point we "fly" down to
	DotWindow w;			//our window
	HSFractal mandelbrot; 	//the 2D-array that will contain all the info
	
	HSMandelbrotWin(int s, double z){
		size = s;
		scale = size/2;
		zoom = z;
		w = new DotWindow(size, size, 1);
		//the fractal array is created, a default middle point is selected
		mandelbrot = new HSFractal(size, scale, -0.5, 0.0);
		//apparently no animation has been chosen, we assign default values to the target point
		animation = new HSPoint(0, 0);
	}
	
	HSMandelbrotWin(int s, double z, HSPoint a){
		size = s;
		scale = size/2;
		zoom = z;
		w = new DotWindow(size, size, 1);
		//animation has been chosen and the target point gets values chosen by the user
		animation = new HSPoint(a.r, a.i);
		//fraktalfltet skapas med anvndarens mittpunkt
		mandelbrot = new HSFractal(size, scale, animation.r , animation.i);
	}
	
	//the main drawing function
	void display(){
		//the fractal array is calculated
		mandelbrot.calculateFractal();
		for(int i = 0; i < size; i++){
			for(int j = 0; j < size; j++){
				w.setDot(i, j, mandelbrot.fractal[i][j].col);
			}
		}
	}
	
	//interactivity, input of values and stop conditions
	void run(){
		//are we to animate or not?
		if(animation.r == 0 && animation.i == 0){
			//if not then we draw the mandelbrot with the default middle point
			display();
			//here we check if the user presses the left mouse button and then releases it
			w.checkMouse(false, false, true, false, false);
			//we check 2 times per second
			w.timeStep(500);
			while(true){
				GameEvent e = w.getNextEvent();
				//did the user press a mouse button
				if(e.getKind() == GameEvent.MOUSE_RELEASED) {
					//we register the point the user pressed
	            	int x = e.getX(); int y = e.getY();
	            	//the scale is lowered by the zoom factor
	            	scale *= zoom;
	            	System.out.println(mandelbrot.fractal[x][y].r);
	            	System.out.println(mandelbrot.fractal[x][y].i);
	            	System.out.println("pixelsize: "+(1/scale));
	            	System.out.println("");
	            	//a new fractal array is created with the help of the window size, scale and the middle point chosen by the user
	            	mandelbrot = new HSFractal(size, scale, mandelbrot.fractal[x][y].r, mandelbrot.fractal[x][y].i);
	            	//the new area gets drawn
	            	display();
	            }
	        }
	     //in case we animate
	    }else{
	    	boolean goOn = true;
	    	int i = 0;
	    	w.checkKey(false, false, true);
			w.timeStep(500);
	    	while(goOn){
	    		GameEvent e = w.getNextEvent();
	    		//a new fractal array is created with the help of the point selected by the user
	    		mandelbrot = new HSFractal(size, scale, animation.r, animation.i);
	    		//the fractal gets calculated
	    		mandelbrot.calculateFractal();
	    		//the color information is saved to file
	    		mandelbrot.saveImage("result"+ i +".png");
	    		//the scale gets lowered with the help of the zoom factor
	    		scale *= zoom;
	    		//the file numbering increases with one
	    		i++;
	    		//in case the scale is lower than 10 raised to 16 or if the user pressed the s-button, in any case, we quit
	    		if(1/scale < 1.0E-19 || e.getKey() == 's')
	    			goOn = false;
	    	}	
	    }
	}
}

/**
 * Main class
 */
class HSMandelbrot{
	public static void main(String[] args) {
		HSPoint start = new HSPoint();;
		int s = Keyboard.nextInt("Window size?: ");
		double z = Keyboard.nextDouble("Window zoom?: ");
		int a = Keyboard.nextInt("Want to animate, yes or no?(input:1 or 0): ");
		if(a == 1){
			start.r = Keyboard.nextDouble("Real starting point?: ");
			start.i = Keyboard.nextDouble("Imaginary starting point?: ");
			//in case we are to animate then we use this constructor
			HSMandelbrotWin MW = new HSMandelbrotWin(s, z, start);
			MW.run();
		}else{
			//if not, then we use this one
			HSMandelbrotWin MW = new HSMandelbrotWin(s, z);
			MW.run();
			
		}	
	}
}